[Setting name="Print debug log statements"]
bool Setting_Debug = false;

namespace UltrawideUiFix {
	const string SHIFTED_BY_DATA_ATTRIBUTE = 'ultrawideuifix-shifted-by';

	const string LAST_VALUE_DATA_ATTRIBUTE = 'ultrawideuifix-last-value';

	const string MODULE_CLASS = 'component-modelibs-uimodules-module';

	const string MODULE_PREFIX = 'UIModule_';

	const string MANIALINK_TAG = '\n<manialink name="';

	const string CONTENT_FRAME_ID = 'frame-content';

	const int MOVE_AUTO = 0;

	const int MOVE_LEFT = -1;

	const int MOVE_RIGHT = 1;

	dictionary MODULES_TO_MOVE = {
		{'Race_BestRaceViewer',							MOVE_AUTO},	// right, best & previous times
		{'Race_Countdown',								MOVE_AUTO},	// right, Time Attack countdown
		{'Race_DisplayMessage',							MOVE_AUTO},	// top left, messages
		{'Race_LapsCounter',							MOVE_AUTO},	// top right, lap counter
		{'Race_Record',									MOVE_AUTO},	// top left, records
		{'Race_RespawnHelper',							MOVE_AUTO},	// right, Respawn & Give up prompts
		{'Race_SpectatorBase_Commands',					MOVE_AUTO},	// bottom right, spectator mode picker
		{'Race_TimeGap',								MOVE_AUTO},	// bottom right, checkpoint times
		{'Race_WarmUp',									MOVE_AUTO},	// right, warmup information
		// COTD
		{'COTDQualifications_QualificationsProgress',	MOVE_AUTO},	// left, qualification information
		{'COTDQualifications_Ranking',					MOVE_AUTO},	// left, qualification results
		{'Knockout_KnockoutInfo',						MOVE_AUTO},	// left, knockout results
		// Royal
		{'Royal_FinishFeed',							MOVE_AUTO},	// top right, recently finished players
		{'Royal_LiveRanking',							MOVE_AUTO},	// top left, team ranking
		{'Royal_RespawnHelper',							MOVE_AUTO},	// right, Restart prompt
		{'Royal_TeamScore',								MOVE_AUTO},	// top right, team results
		// Ranked
		{'Teams_Matchmaking_LiveRanking',				MOVE_LEFT}	// left, player positions and points
	};

	dictionary CLIPS_TO_EXPAND = {
		// COTD
		{
			'COTDQualifications_QualificationsProgress', array<string> = {
				'frame-global' // left, qualification information
			}
		},
		{
			'COTDQualifications_Ranking', array<string> = {
				'frame-global' // left, qualification results
			}
		}
	};

	namespace Fixes {
		void fixModule(const string &in moduleId, float uiShift, CGameManialinkControl@ control) {
			// Check if we want to modify this module
			if (MODULES_TO_MOVE.Exists(moduleId)) {
				if (control.DataAttributeGet(LAST_VALUE_DATA_ATTRIBUTE) == tostring(control.PosnX) && control.DataAttributeGet(SHIFTED_BY_DATA_ATTRIBUTE) == tostring(uiShift)) {
					// If our position was kept and the module has already been moved by our current shift value, we don't want to do anything
					if (Setting_Debug) {
						trace('Module ' + moduleId + ' has already been moved with this shift value and the position has stayed the same, ignoring');
					}
				} else {
					// By default, start shifting from the initial module position
					float prevShift = 0.0;

					// If our position was kept, that means the shift value has changed, get the previous shift value and use it
					if (control.DataAttributeGet(LAST_VALUE_DATA_ATTRIBUTE) == tostring(control.PosnX)) {
						prevShift = Text::ParseFloat(control.DataAttributeGet(SHIFTED_BY_DATA_ATTRIBUTE));
					} else if (Setting_Debug) {
						trace('Module ' + moduleId + ' has had its position changed by someone else, ignoring our previous shift value');
					}

					// Get the direction we want to shift in
					int direction = int(MODULES_TO_MOVE[moduleId]);
					if (direction == MOVE_AUTO) {
						direction = control.PosnX >= 0 ? 1 : -1;
					}

					// Calculate the shift and move the module
					float controlShift = direction * (uiShift - prevShift);
					control.PosnX += controlShift;
					print('Module ' + moduleId + ' moved by ' + controlShift + ' units');

					if (Setting_Debug) {
						trace('Module ' + moduleId + ' has been moved from X=' + (control.PosnX - controlShift) + ' to X=' + control.PosnX);
					}

					// Mark the module as modified and store the last position
					control.DataAttributeSet(SHIFTED_BY_DATA_ATTRIBUTE, tostring(uiShift));
					control.DataAttributeSet(LAST_VALUE_DATA_ATTRIBUTE, tostring(control.PosnX));
				}
			}
		}

		void fixClips(const string &in moduleId, float uiShift, CGameManialinkPage@ page) {
			// Check if we have any clips to modify for this module
			if (CLIPS_TO_EXPAND.Exists(moduleId)) {
				string[] clips = cast<array<string>>(CLIPS_TO_EXPAND[moduleId]);

				for (uint clipIndex = 0; clipIndex < clips.Length; ++clipIndex) {
					// Check if a control exists for every clip ID provided for the given module
					CGameManialinkControl@ clip = page.GetFirstChild(clips[clipIndex]);

					if (clip is null) {
						continue;
					}

					string clipId = clip.ControlId;

					if (clip.DataAttributeGet(LAST_VALUE_DATA_ATTRIBUTE) == tostring(clip.Size.x) && clip.DataAttributeGet(SHIFTED_BY_DATA_ATTRIBUTE) == tostring(uiShift)) {
						// If our size was kept and the module has already been resized by our current shift value, we don't want to do anything
						if (Setting_Debug) {
							trace('Clip ' + clipId + ' has already been expanded with this shift value and the size has stayed the same, ignoring');
						}
					} else {
						// By default, start shifting from the initial clip size
						float prevShift = 0.0;

						// If our size was kept, that means the shift value has changed, get the previous shift value and use it
						if (clip.DataAttributeGet(LAST_VALUE_DATA_ATTRIBUTE) == tostring(clip.Size.x)) {
							prevShift = Text::ParseFloat(clip.DataAttributeGet(SHIFTED_BY_DATA_ATTRIBUTE));
						} else if (Setting_Debug) {
							trace('Clip ' + clipId + ' has had its size changed by someone else, ignoring our previous shift value');
						}

						// Calculate the expansion and expand the clip
						float clipExpand = 2 * (uiShift - prevShift);
						clip.Size = vec2(clip.Size.x + clipExpand, clip.Size.y);
						print('Clip ' + clipId + ' expanded by ' + clipExpand + ' units');

						if (Setting_Debug) {
							trace('Clip ' + clipId + ' has been expanded from W=' + (clip.Size.x - clipExpand) + ' to W=' + clip.Size.x);
						}

						// Mark the module as modified and store the last size
						clip.DataAttributeSet(SHIFTED_BY_DATA_ATTRIBUTE, tostring(uiShift));
						clip.DataAttributeSet(LAST_VALUE_DATA_ATTRIBUTE, tostring(clip.Size.x));
					}
				}
			}
		}
	}

	class Fixer {
		// Whether our fixes should be applied during the next frame, used to make sure layer data is available
		bool applyNextFrame = false;

		// Whether our fixes should be applied at a specific timestamp in the future
		uint applyAtTimestamp = 0;

		// The amount of layers that existed in the previous frame, used to make sure we only apply fixes when the UI changes
		uint previousLayerLength = 0;

		// The map UID for the previous frame, used to make sure we only apply fixes when the UI changes
		string previousMapUid;

		// The US shift for the previous frame, used to make sure we only apply fixes when the UI changes
		float previousUiShift = 0.;

		// Get the amount of units to shift UI elements by (https://maniaplanet-community.gitbook.io/maniascript/manialink/manialinks#manialink-positioning-system)
		float uiShift {
			get {
				float h = Draw::GetHeight();

				if (h == 0) {
					return 0;
				} else {
					return Math::Max(0.0, (float(Draw::GetWidth()) / (h / 9)) * 10 - 160);
				}
			}
		}

		// Normalize UI layer buffer
		MwFastBuffer<CGameUILayer@> layers {
			get const {
				CGameManiaAppPlayground@ playground = GetApp().Network.ClientManiaAppPlayground;

				// Check if layers are available
				if (playground !is null) {
					return playground.UILayers;
				} else {
					// Layers are not yet available, use an empty buffer
					MwFastBuffer<CGameUILayer@> no_layers;
					return no_layers;
				}
			}
		};

		// Get the current map UID
		string GetMapUid() {
			CGameCtnApp@ app = GetApp();

			if (app.CurrentPlayground is null) {
				return '';
			} else if (app.RootMap !is null) {
				return app.RootMap.MapInfo.MapUid;
			} else {
				return '';
			}
		}

		// Check if the UI has changed since the last call
		bool HasUiChanged() {
			if (Setting_Debug) {
				if (layers.Length != previousLayerLength) {
					trace('Layer count changed from ' + previousLayerLength + ' to ' + layers.Length);
				}

				if (previousUiShift != uiShift) {
					trace('UI shift changed from ' + previousUiShift + ' to ' + uiShift);
				}

				if (previousMapUid != GetMapUid()) {
					trace('Map UID changed from ' + previousMapUid + ' to ' + GetMapUid());
				}
			}

			return layers.Length != previousLayerLength || previousUiShift != uiShift || previousMapUid != GetMapUid();
		}

		// Apply our position and size fixes
		void ApplyFixes() {
			// Loop through each layer
			for (uint layerIndex = 0; layerIndex < layers.Length; ++layerIndex) {
				CGameUILayer@ layer = layers[layerIndex];
				CGameManialinkPage@ page = layer.LocalPage;

				CGameManialinkControl@ control = null;
				string moduleId;

				// Check if we have the main UI module
				page.GetClassChildren(MODULE_CLASS, page.MainFrame, true);

				if (page.GetClassChildren_Result.Length > 0) {
					@control = page.GetClassChildren_Result[0];
					moduleId = control.ControlId;
				} else {
					// If not, check if we have a manialink tag with a name attribute
					string manialinkSnippet = layer.ManialinkPageUtf8.SubStr(0, 100);

					if (manialinkSnippet.StartsWith(MANIALINK_TAG)) {
						// If we do have one, get the value of the attribute and check if it has the UI module prefix
						string[] manialinkSnippetParts = manialinkSnippet.Split('"');

						if (manialinkSnippetParts.Length > 1) {
							string manialinkName = manialinkSnippetParts[1];

							if (manialinkName.StartsWith(MODULE_PREFIX)) {
								// If it does strip it and keep the rest
								moduleId = manialinkName.SubStr(MODULE_PREFIX.Length);

								// Check if we have a frame-content frame, if we do, use it as the control
								@control = page.GetFirstChild(CONTENT_FRAME_ID);

								if (control is null) {
									if (Setting_Debug) {
										trace('Unable to find content frame in layer ' + moduleId);
									}
									continue;
								}
							}
						}
					}
				}

				if (moduleId == '' || control is null) {
					if (Setting_Debug) {
						trace('Unable to identify layer ' + layer.IdName);
					}
					continue;
				}

				// Move the main UI module using properties
				Fixes::fixModule(moduleId, uiShift, @control);

				// Resize clips using properties
				Fixes::fixClips(moduleId, uiShift, @page);
			}
		}

		// Store state so we have something to compare to when we get to the next frame
		void UpdateState() {
			previousLayerLength = layers.Length;
			previousUiShift = uiShift;
			previousMapUid = GetMapUid();
		}
	}
}

// The main instance of our logic class
UltrawideUiFix::Fixer@ fixer = null;

// Main entry point
void Main() {
	// Instantiate our logic class
	UltrawideUiFix::Fixer mainFixer;

	// Store the instance as the main one
	@fixer = @mainFixer;
}

// Render function called every frame
void Render() {
	// Check if our fixer instance is available
	if (fixer !is null) {
		// First, check if we should apply fixes based on the previous frame
		if (fixer.applyNextFrame == true) {
			fixer.ApplyFixes();
			fixer.applyNextFrame = false;
		}

		// Check if there were any UI changes, if so, schedule applying our fixes for the next frame and a second in the future
		if (fixer.HasUiChanged()) {
			fixer.applyNextFrame = true;

			if (Setting_Debug) {
				trace('UI change detected, applying fixes');
			}

			// As some UI changes apply in Maniascript init functions, also schedule applying our fixes one second in the future to catch them
			fixer.applyAtTimestamp = Time::get_Now() + 1000;

			if (Setting_Debug) {
				trace('Scheduling fix application in 1 second');
			}
		}

		// Check if we've scheduled applying our fixes at a specific timestamp, if so, schedule applying our fixes for the next frame and clear the timestamp
		if (fixer.applyAtTimestamp != 0 && Time::get_Now() > fixer.applyAtTimestamp) {
			fixer.applyNextFrame = true;
			fixer.applyAtTimestamp = 0;

			if (Setting_Debug) {
				trace('Scheduled application reached, applying fixes');
			}
		}

		// Update fixer state so we can check for changes in the next frame
		fixer.UpdateState();
	}
}
